/*
 *
 * dbridge_db.c
 * Description:
 * DOCSIS bridge database implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#define _DBRIDGE_DB_C_

/*! \file dbridge_db.c
    \brief the docsis bridge database
*/

/**************************************************************************/
/*      INCLUDES:                                                         */
/**************************************************************************/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/list.h>
#include <linux/ti_hil.h>

#include "dbridge_common.h"
#include "dbridge_db.h"
#include "dbridge_hal_filter.h"
#include "dbridge_igmp.h"
#include "dbridge_l2vpn_ds_db.h"
#include "dbridge_mcast.h"
#include "dbridge_mdf_db.h"
#include "dbridge_netdev.h"
#include "pal.h"

/**************************************************************************/
/*      EXTERNS Declaration:                                              */
/**************************************************************************/

/**************************************************************************/
/*      DEFINES:                                                          */
/**************************************************************************/

/**************************************************************************/
/*      LOCAL DECLARATIONS:                                               */
/**************************************************************************/

#define MIN_NUM_CPE 1 

/**************************************************************************/
/*      LOCAL VARIABLES:                                                  */
/**************************************************************************/

/*  Structure which keeps track of all the information for the DOCSIS Bridge. */
DbridgeDataBase_t dbridge_db;
static DbridgeCounters_t  dbridge_counters; 
static unsigned  int devices_table_insert;
static unsigned  int alt_table_insert;
static unsigned  int data_base_init = 0;
static unsigned  int max_cpe = MIN_NUM_CPE;
static unsigned  int lbridge_interface_mask;

static void DbrigdeDb_AltDelEntry(DbridgeAltEntry_t *altEntry);

/**************************************************************************/
/*      INTERFACE FUNCTIONS Implementation:                               */
/**************************************************************************/

/**************************************************************************/
/*      LOCAL FUNCTIONS:                                                  */
/**************************************************************************/
#ifdef CONFIG_TI_DOCSIS_EGRESS_HOOK
static int configure_docsis_egress_hook_handler(int reg);
static int Dbridge_egress_hook(struct sk_buff* skb);
static struct net_device *usb0_netdev;
static struct net_device *eth0_netdev;

#endif /* CONFIG_TI_DOCSIS_EGRESS_HOOK */

/**************************************************************************/
/*! \fn int DbridgeDb_Init(void)                                     
 **************************************************************************
 *  \brief DOCSIS bridge data base initialization.
 *  \param[in] no input.
 *  \param[out] no output.
 *  \return OK or error status.
 **************************************************************************/
int DbridgeDb_Init(void)
{
    int i;

    struct net_device *eth_dev =dev_get_by_name (ETH_DEV_NAME);
    struct net_device *usb_dev =dev_get_by_name (USB_DEV_NAME);
    memset ((void *)&dbridge_db, 0, sizeof(DbridgeDataBase_t));
    memset ((void *)&dbridge_counters, 0, sizeof(DbridgeCounters_t));

    devices_table_insert=0;
    alt_table_insert=0;
    data_base_init=1;

    /* Put all ALT entries on the FREE list */
    INIT_LIST_HEAD(&DBRALT_FREE_HEAD);
    INIT_LIST_HEAD(&DBRALT_OCCUPIED_HEAD);
    for (i = 0; i < DBR_MAX_ALT_SIZE; i++)
    {
        /* Init cpe num, only DBR_MAX_CPE_ENTRY, all the rest get 0 */
        if (i >= (DBR_MAX_ALT_SIZE - DBR_MAX_CPE_ENTRY))
        {
            /* cpe num [1..DBR_MAX_CPE_ENTRY] */
            dbridge_db.alt.altTable[i].cpe_number = i - (DBR_MAX_ALT_SIZE - DBR_MAX_CPE_ENTRY) + 1;
        }
        else
        {
            /* 0 (the memset above does the trick, but this is cleaner...) */
            dbridge_db.alt.altTable[i].cpe_number = 0;
        }
        /* Make the initial order: */
        /* - Group of cpe_num == 0 */
        /* - cpe_num == 1 */
        /* - cpe_num == 2 */
        /* - cpe_num == ... */
        /* - cpe_num == DBR_MAX_CPE_ENTRY */
        /* This is done by add_tail, which results in 1st element 1st */
        list_add_tail(&dbridge_db.alt.altTable[i].listAltEntry, &DBRALT_FREE_HEAD);
    }

    /* This is patch for filtering need to be change when using usb0 */
    if (eth_dev){
        lbridge_interface_mask = ((unsigned long) 1 << ((eth_dev->ifindex)-1));
        dev_put(eth_dev);
    }
    if (usb_dev){
        lbridge_interface_mask |= ((unsigned long) 1 << ((usb_dev->ifindex)-1));
        dev_put(usb_dev);
    }
    return 0; /* OK*/
}

/**************************************************************************/
/*! \fn int DbridgeDb_ReInit(void)                                     
 **************************************************************************
 *  \brief DOCSIS bridge data base re-initialization.
 *  \param[in] no input.
 *  \param[out] no output.
 *  \return OK or error status.
 **************************************************************************/
int DbridgeDb_ReInit(void)
{
    DbridgeAltEntry_t *altEntry;
    DbridgeAltEntry_t *next = NULL;
    unsigned int cpe_type;
    DBRALT_PROT_DCL(lockKey);

    /* Protect alt_table_insert */
	DBRALT_PROT_ON(lockKey);

    /* remove all cpe from alt table */
    list_for_each_entry_safe(altEntry, next, &DBRALT_OCCUPIED_HEAD, listAltEntry)
    {
        cpe_type = altEntry->macLearningFrom;
        if ((cpe_type == CPE_STATIC) || (cpe_type == CPE_DYNAMIC))
        {
            /* Remove HW filter for unicast MAC address */   
            HAL_MacDsUnicastDataFilterDelete(altEntry->macAddress);

            DbrigdeDb_AltDelEntry(altEntry);
        }
    }

	DBRALT_PROT_OFF(lockKey);

    dbridge_counters.StaticCpe = 0;
    dbridge_counters.DynamicCpe = 0;
    max_cpe = MIN_NUM_CPE;

    /* remove all HW filters */

    /* remove all multicast filters */
    DbridgeDb_DelAllMcastMac();

    return 0; /* OK*/
}

/**************************************************************************/
/*! \fn DDbridgeAltEntry_t* DbridgeDb_AltGetMemberByMac(unsigned char* mac_address)
 **************************************************************************
 *  \brief The function retreives the dbridge ALT table member for the  
 *          corresponding mac address.
 *      ****   NOTE ****
 *          Protection must be supplied by the caller
 *          Protection may end when the output pointer is no longer in use
 *  \param[in] cpe mac address.
 *  \return Success-pointer to the Dbridge ALT table, NULL-If no entry is found.
 **************************************************************************/
DbridgeAltEntry_t* DbridgeDb_AltGetMemberByMac(unsigned char* mac_address)
{
    DbridgeAltEntry_t *altEntry = NULL;
    DbridgeAltEntry_t *altResult = NULL;

    DBRALT_FOREACH_ALT_ENTRY(altEntry)
    {
        if (memcmp(mac_address, altEntry->macAddress, MAC_ADDRESS_LEN) == 0)
        {
            altResult = altEntry;
            break;
        }
    }

    return altResult;
}

/**************************************************************************/
/*! \fn int DbridgeDb_AltAddMac(unsigned char* mac_address, Maclearning_e mac_learn_from,struct net_device *ptr_netdev, struct net_device*  sourceInterface)
 **************************************************************************
 *  \brief this function Add MAC address to ALT table or update device information for
 *  \brif existing cpe MAC that add from config file .
 *  \param[in] cpe mac address.
 *  \param[in] mac address from - NOT_CPE, CPE_STATIC or CPE_DYNAMIC,
 *  \param[in] pointer to docsis device table.
 *  \param[in] pointer to first interface where the packet received from.
 *  \param[out] no output.
 *  \return OK or error status.
 **************************************************************************/
int DbridgeDb_AltAddMac(unsigned char* mac_address, Maclearning_e mac_learn_from, 
                        DbridgeDevice_t *dbridge_Devices,struct net_device*  sourceInterface)
{   
    DbridgeAltEntry_t* search_dbridge_alt; 
    DbridgeAltEntry_t* ptr_dbridge_alt; 
    DBRALT_PROT_DCL(lockKey);
    int retval;
    int st = 1;

	DBRALT_PROT_ON(lockKey);
    ptr_dbridge_alt = DbridgeDb_AltGetMemberByMac(mac_address);
    if (ptr_dbridge_alt!=NULL)
    {
        /* MAC address exists update the corresponding device */
        if (mac_learn_from == CPE_DYNAMIC && ptr_dbridge_alt->dbridgeDevice==NULL)
        {
            ptr_dbridge_alt->dbridgeDevice= dbridge_Devices;
            ptr_dbridge_alt->sourceInterface = sourceInterface;

            printk(KERN_INFO "Update source from %.2x-%.2x-%.2x-%.2x-%.2x-%.2x %s\n",
                   (unsigned char)mac_address[0],
                   (unsigned char)mac_address[1],
                   (unsigned char)mac_address[2],
                   (unsigned char)mac_address[3],
                   (unsigned char)mac_address[4],
                   (unsigned char)mac_address[5],
                   sourceInterface->name);
        }

        DBRALT_PROT_OFF(lockKey);

        return 0; 
    }

    /* check if we have place for one more cpe */
    if ((mac_learn_from == CPE_DYNAMIC) || (mac_learn_from == CPE_STATIC))
    {
        if ((dbridge_counters.StaticCpe + dbridge_counters.DynamicCpe ) >= max_cpe)
        {
            DBRALT_PROT_OFF(lockKey);

            /* no place for more CPEs */
            printk(KERN_INFO "No room for %s ( static %d, dynamic %d,  max %d) with MAC %.2x-%.2x-%.2x-%.2x-%.2x-%.2x\n",
                   (mac_learn_from == CPE_DYNAMIC) ? "CPE_DYNAMIC" : "CPE_STATIC",
                   dbridge_counters.StaticCpe, dbridge_counters.DynamicCpe, max_cpe,
                   mac_address[0], mac_address[1], mac_address[2], mac_address[3], mac_address[4], mac_address[5]);
            return -1;
        }
    }

    /* Create new, keep protection on from above */
    ptr_dbridge_alt = NULL;
    /* Get free element */
    list_for_each_entry(search_dbridge_alt, &DBRALT_FREE_HEAD, listAltEntry)
    {
        /* Check if this is a "good" entry. */
        if (/* For CPE entries, we need cpe_num > 0, < MAX */
            (((mac_learn_from == CPE_DYNAMIC) || (mac_learn_from == CPE_STATIC)) &&
             ((search_dbridge_alt->cpe_number <= 0) || (search_dbridge_alt->cpe_number > DBR_MAX_CPE_ENTRY))) ||
            /* For non-CPE entries, we need cpe_num == 0 */
            ((mac_learn_from == DBRIDGE_INIT) && (search_dbridge_alt->cpe_number != 0)) )
        {
            /* Not good, try next */
            continue;
        }
        else
        {
            /* Found */
            ptr_dbridge_alt = search_dbridge_alt;
            break;
        }
    }
    /* Did we find ? */
    if (ptr_dbridge_alt != NULL)
    {
        /* Found */
        if (mac_learn_from==CPE_DYNAMIC)
            dbridge_counters.DynamicCpe++;
        else if (mac_learn_from==CPE_STATIC)
            dbridge_counters.StaticCpe++;

        memcpy (ptr_dbridge_alt->macAddress, mac_address, MAC_ADDRESS_LEN);
        /* add ptr to docsis device */
        ptr_dbridge_alt->dbridgeDevice = dbridge_Devices;
        ptr_dbridge_alt->sourceInterface = sourceInterface;
        ptr_dbridge_alt->macLearningFrom=mac_learn_from;

        if (DBRIDGE_INIT != mac_learn_from)
        {
            /* open HW filter for unicast MAC address */   
            st = HAL_MacDsUnicastDataFilterAdd(mac_address);
        }

        /* Add at end, so the INIT types alway (?) appear 1st */
        list_move_tail(&ptr_dbridge_alt->listAltEntry, &DBRALT_OCCUPIED_HEAD);

        retval = 0;
    }

    DBRALT_PROT_OFF(lockKey);

    /* Protection is OFF, do not use ptr_dbridge_alt */
    /* Can stil check if it was NULL, but do not use it... */
    if (ptr_dbridge_alt == NULL)
    {
        /* Not found */
        printk(KERN_ERR "Could not add a new ALT entry for MAC %.2x-%.2x-%.2x-%.2x-%.2x-%.2x\n",
               mac_address[0], mac_address[1], mac_address[2],mac_address[3], mac_address[4], mac_address[5]);
        retval = -1;
    }
    else
    {
        /* Found */
        printk(KERN_INFO "Add to ALT (with%s source filter) %.2x-%.2x-%.2x-%.2x-%.2x-%.2x %s\n",
               st == 0 ? "" : "out",
               mac_address[0], mac_address[1], mac_address[2],mac_address[3], mac_address[4], mac_address[5],
               sourceInterface == NULL ? "noName" : sourceInterface->name);
        retval = 0;
    }

    return retval;
}

/**************************************************************************/
/*! \fn int DbridgeDb_AltDelMac(unsigned char* mac_address, 
            Maclearning_e mac_learn_from, char *dest_dev_name, char *src_dev_name, 
            unsigned char cpe_number, int net_dev_type,, DbridgeDb_FieldMask_e fieldMask)
 **************************************************************************
 *  \brief DEL MAC address from ALT table.
 *          All ALT entries are compared against the input params selected by the mask.
 *          Depending on DBRALT_MASK_ALL/DBRALT_MASK_ANY, entries that match all/any input
 *          are deleted.
 *  \param[in] cpe mac address.
 *  \param[in] mac address from - DBRIDGE_INIT, CPE_STATIC or CPE_DYNAMIC,
 *  \param[in] dest_dev_name - name of destination device
 *  \param[in] src_dev_name - name of source device
 *  \param[in] cpe_number
 *  \param[in] net_dev_type - device type, e.g. DBR_ESAFE_NET_DEV_TYPE, DBR_CMCI_NET_DEV_TYPE
 *  \param[in] field_mask - Which fields are used for deciding which CPEs to del
 *  \param[out] no output.
 *  \return OK or error status.
 **************************************************************************/
int DbridgeDb_AltDelMac(unsigned char* mac_address,
                        Maclearning_e mac_learn_from, 
                        char *dest_dev_name,
                        char *src_dev_name,
                        unsigned char cpe_number,
                        int net_dev_type,
                        DbridgeDb_FieldMask_e field_mask)
{   
    DbridgeAltEntry_t* ptr_dbridge_alt; 
    DbridgeAltEntry_t *next = NULL;
    DBRALT_PROT_DCL(lockKey);
    DbridgeDb_FieldMask_e actualMatch;

#define DBRALT_DEL_TEST_MATCH(__field, __match) \
    do \
    { \
        if (field_mask & __field) \
        { \
            if (__match) \
            { \
                /* Found match */ \
                if (field_mask & DBRALT_MASK_ANY) \
                { \
                    /* 1 match is enough, del & continue to next entry */ \
                    printk(KERN_INFO "DEL ALT %.2x-%.2x-%.2x-%.2x-%.2x-%.2x, match ANY && %s\n", \
                            ptr_dbridge_alt->macAddress[0], ptr_dbridge_alt->macAddress[1], ptr_dbridge_alt->macAddress[2], \
                            ptr_dbridge_alt->macAddress[3], ptr_dbridge_alt->macAddress[4], ptr_dbridge_alt->macAddress[5], \
                            #__match); \
                    DbrigdeDb_AltDelEntry(ptr_dbridge_alt); \
                    continue; \
                } \
                else \
                { \
                    /* Record match */ \
                    actualMatch |= __field; \
                } \
            } \
            else if (field_mask & DBRALT_MASK_ALL) \
            { \
                /* Single mismatch is enough to continue to next entry */ \
                continue; \
            } \
        } \
    } while (0)

    if ( ((field_mask & DBRALT_MASK_ALL) == 0) && ((field_mask & DBRALT_MASK_ANY) == 0) )
    {
        printk(KERN_ERR "Mask (0x%08X) does not specify neither ANY nor ALL\n", field_mask);
        return -1;
    }

	DBRALT_PROT_ON(lockKey);
    list_for_each_entry_safe(ptr_dbridge_alt, next, &DBRALT_OCCUPIED_HEAD, listAltEntry)
    {
        if (field_mask & DBRALT_MASK_ALL)
        {
            actualMatch = DBRALT_MASK_ALL;
        }
        else
        {
            actualMatch = 0;
        }

        DBRALT_DEL_TEST_MATCH(DBRALT_MASK_MAC, 
                              (mac_address != NULL) && 
                              (memcmp(mac_address, ptr_dbridge_alt->macAddress, MAC_ADDRESS_LEN) == 0) );
        DBRALT_DEL_TEST_MATCH(DBRALT_MASK_LEARNFROM, mac_learn_from == ptr_dbridge_alt->macLearningFrom);
        DBRALT_DEL_TEST_MATCH(DBRALT_MASK_DESTDEV, 
                              (dest_dev_name != NULL) && 
                              (ptr_dbridge_alt->dbridgeDevice != NULL) && 
                              (ptr_dbridge_alt->dbridgeDevice->ptr_interface != NULL) && 
                              (ptr_dbridge_alt->dbridgeDevice->ptr_interface->name != NULL) && 
                              strcmp(dest_dev_name, ptr_dbridge_alt->dbridgeDevice->ptr_interface->name) == 0);
        DBRALT_DEL_TEST_MATCH(DBRALT_MASK_SRCDEV, 
                              (src_dev_name != NULL) && 
                              (ptr_dbridge_alt->sourceInterface != NULL) && 
                              (ptr_dbridge_alt->sourceInterface->name != NULL) && 
                              strcmp(src_dev_name, ptr_dbridge_alt->sourceInterface->name) == 0);
        DBRALT_DEL_TEST_MATCH(DBRALT_MASK_CPENUM, cpe_number == ptr_dbridge_alt->cpe_number);
        DBRALT_DEL_TEST_MATCH(DBRALT_MASK_DEVTYPE, 
                              (ptr_dbridge_alt->dbridgeDevice != NULL) &&
                              (ptr_dbridge_alt->dbridgeDevice->net_dev_type == net_dev_type));

        if ((field_mask & DBRALT_MASK_ALL) && (actualMatch == field_mask))
        {
            /* ALL && all fields match ==> DEL */
            printk(KERN_INFO "DEL ALT %.2x-%.2x-%.2x-%.2x-%.2x-%.2x, match ALL", 
                    ptr_dbridge_alt->macAddress[0], ptr_dbridge_alt->macAddress[1], ptr_dbridge_alt->macAddress[2], 
                    ptr_dbridge_alt->macAddress[3], ptr_dbridge_alt->macAddress[4], ptr_dbridge_alt->macAddress[5]);
            if (field_mask & DBRALT_MASK_MAC) printk(" && DBRALT_MASK_MAC");
            if (field_mask & DBRALT_MASK_LEARNFROM) printk(" && DBRALT_MASK_LEARNFROM");
            if (field_mask & DBRALT_MASK_DESTDEV) printk(" && DBRALT_MASK_DESTDEV");
            if (field_mask & DBRALT_MASK_SRCDEV) printk(" && DBRALT_MASK_SRCDEV");
            if (field_mask & DBRALT_MASK_CPENUM) printk(" && DBRALT_MASK_CPENUM");
            if (field_mask & DBRALT_MASK_DEVTYPE) printk(" && DBRALT_MASK_DEVTYPE");
            printk("\n");

            DbrigdeDb_AltDelEntry(ptr_dbridge_alt);
        }
    }
    DBRALT_PROT_OFF(lockKey);

    return 0;
}

/**************************************************************************/
/*! \fn int DbridgeDb_AltAddMacByDevName(unsigned char* mac_address, unsigned char* name)
 **************************************************************************
 *  \brief this function Add MAC address to ALT table for cm interface (wan0 or lan0)
 *  \param[in] mac address.
 *  \param[in] device name.
 *  \param[out] no output.
 *  \return OK or error status.
 **************************************************************************/
int DbridgeDb_AltAddMacByDevName(unsigned char* mac_address, unsigned char* name)
{
    int index;
    DbridgeDevice_t*   ptr_dbridge_dev;
    int name_len = strlen(name);

    for (index=0 ; index<devices_table_insert; index++)
    {
        ptr_dbridge_dev = &dbridge_db.DbridgeDevices[index];
        if ((strlen(ptr_dbridge_dev->ptr_interface->name)== name_len) &&
            (memcmp ((void *)name, (void *)&ptr_dbridge_dev->ptr_interface->name, name_len)==0))
        {
            /* find match between inpute name and dbridge device table */
            /* add the the mac information to ALT table */ 
            if (DbridgeDb_AltAddMac(mac_address,DBRIDGE_INIT, ptr_dbridge_dev,NULL)==0)
                return 0;
            else
            {
                return -1;
            }
        }
    }
    printk(KERN_ERR "Could not find %s with MAC %.2x-%.2x-%.2x-%.2x-%.2x-%.2x in DbrDB Devices\n",
           name, mac_address[0], mac_address[1], mac_address[2], mac_address[3], mac_address[4], mac_address[5]);
    return -1;
}

/***************************************************************************/
/*! \fn int DbridgeDB_SetMaxCpe(unsigned  int mac_cpe_config)                                     
 **************************************************************************
 *  \brief set maximum  supported CPEs.
 *  \param[in] Max allow CPEs .
 *  \param[out] no output.
 *  \return OK or error status.
 **************************************************************************/
int DbridgeDB_SetMaxCpe(unsigned  int mac_cpe)
{
    max_cpe = mac_cpe;
    return 0;   /* OK*/
}

/***************************************************************************/
/*! \fn int DbridgeDB_GetMaxCpe()                                     
 **************************************************************************
 *  \brief Get maximum  supported CPEs.
 *  \param[in] None .
 *  \param[out] None.
 *  \return Max CPEs.
 **************************************************************************/
int DbridgeDB_GetMaxCpe(void)
{
    return max_cpe; 
}

/***************************************************************************/
/*! \fn int DbridgeDb_DevAdd(struct net_device* ptr_net_dev,int (*push)(struct sk_buff* skb, 
            unsigned int fc_flags))
 ***************************************************************************
 *  \brief Add member to device table.
 *  \param[in] pointer to net device.
 *  \param[in] The PUSH Function to send the packet to Interface.
 *  \param[in] docsis interface type  - CPE eSAFE CABLE WAN IP or LAN IP.
 *  \return 0=OK or error status.
 **************************************************************************/
int DbridgeDb_DevAdd(struct net_device* ptr_net_dev,int (*push)(struct sk_buff* skb, 
                                                                unsigned int fc_flags), int net_dev_type)
{
    if (devices_table_insert < DBR_MAX_INTERFACES)
    {
        dbridge_db.DbridgeDevices[devices_table_insert].push = push;
        dbridge_db.DbridgeDevices[devices_table_insert].ptr_interface = ptr_net_dev;
        dbridge_db.DbridgeDevices[devices_table_insert].net_dev_type = net_dev_type; 
        devices_table_insert++;
        return 0;
    }
    else
    {
        printk(KERN_WARNING "Device table is full\n");

        return -1;
    }

}

/*************************************************************************************/
/*! \fn DbridgeDevice_t* DbridgeDb_DevGetByInterface (struct net_device* ptr_interface)
 *************************************************************************************
 *  \brief The function retreives the dbridge device table member for the corresponding 
 *  \brief network interface.
 *  \param[in] pointer to net net device.
 *  \return Success-pointer to the Dbridge device table, NULL-If no entry is found.
 *************************************************************************************/
DbridgeDevice_t* DbridgeDb_DevGetByInterface (struct net_device* ptr_interface)
{
    int index;

    /* Search the DOCSIS Network device table. */
    for (index = 0; index <devices_table_insert; index++)
    {
        /* Compare the network interface for a match. */
        if (ptr_interface == dbridge_db.DbridgeDevices[index].ptr_interface)
            return &dbridge_db.DbridgeDevices[index];
    }

    /* No match found. */
    return NULL;
}

/*************************************************************************************/
/*! \fn DbridgeDevice_t* DbridgeDb_DevGetByIndex (struct net_device* ptr_interface)
 *************************************************************************************
 *  \brief The function retreives the dbridge device table member for the corresponding 
 *  \brief index.
 *  \param[in] index.
 *  \return Success-pointer to the Dbridge device table, NULL-If no entry is found.
 *************************************************************************************/
DbridgeDevice_t* DbridgeDb_DevGetByIndex (int index)
{
    if (index >= 0 && index < devices_table_insert)
    {
        return &dbridge_db.DbridgeDevices[index];
    }
    else
    {
        return NULL;
    }
}

/*************************************************************************************/
/*! \fn DDbridgeDevice_t* DbridgeDb_DevGetByType(int net_dev_type)
 *************************************************************************************
 *  \brief The function retreives the dbridge device table member for the corresponding 
 *  \brief type.
 *  \param[in] net dev type - CMCI, ESAFE CABLE WAN_IP or LAN_IP.
 *  \return Success-pointer to the Dbridge device table, NULL-If no entry is found.
 *************************************************************************************/
DbridgeDevice_t* DbridgeDb_DevGetByType(int net_dev_type)
{
    int index;

    /* Search the DOCSIS Network device table. */
    for (index = 0; index <devices_table_insert; index++)
    {
        /* Compare the network interface for a match. */
        if (net_dev_type == dbridge_db.DbridgeDevices[index].net_dev_type)
            return &dbridge_db.DbridgeDevices[index];
    }
    /* No match found. */
    return NULL;

}

/*************************************************************************************/
/*! \fn int DbridgeDb_GetNumOfDevice (void)
 *************************************************************************************
 *  \brief The function retreives the number of device connect to docsis bridge
 *  \param[in] none.
 *  \return number of device connect to docsis bridge.
 *
 *************************************************************************************/
int DbridgeDb_GetNumOfDevice (void)
{
    return devices_table_insert;
}

/*************************************************************************************/
/*! \fn void DbridgeDb_SetMode (int mode)
 *************************************************************************************
 *  \brief set docsis bridge mode
 *  \param[in] docsis bridge mode - off, registered pre_registration or naco_off 
 *  \return none.
 *************************************************************************************/
void DbridgeDb_SetMode(int mode)
{
    unsigned int mdfMode;
    int erouter_mode;
    static int selective_packet_handler = 0;
    static int register_docsis_egress_hook = 0;
    extern int ti_register_selective_packet_handler (char *br_name, int (*packet_handler)(struct sk_buff *skb), int priority);
    extern int ti_unregister_selective_packet_handler (char *br_name, int (*packet_handler)(struct sk_buff *skb), int priority);

    erouter_mode = DbridgeDb_Erouter_GetMode();


    if ((dbridge_db.mode==DBR_REGISTERED_MODE)&&((mode == DBR_PRE_REGISTRATION_MODE) || (mode == DBR_NACO_OFF_MODE)))
    {
        DbridgeDb_ReInit();
        DbridgeMdfDb_ReInit();
        DbridgeL2vpnDsDb_Init();
    }

    else if (mode == DBR_PRE_REGISTRATION_MODE)
    {
        DbridgeMdfDb_ReInit();
        DbridgeL2vpnDsDb_Init();
    }

    /* Set IGMP mode */
    /* need to move the ti_register_selective_packet_handler to deferent place */
    if (mode == DBR_REGISTERED_MODE)
    {
        if(DISABLE == erouter_mode)
        {
            ti_register_selective_packet_handler (LBRIGE_CONNECTION_DEV_NAME, ti_selective_forward, 1);
            selective_packet_handler = 1;
            configure_docsis_egress_hook_handler(1);
            register_docsis_egress_hook = 1;

        }
        /* if MDF is disabled - start the igmp module */
        DbridgeMdfDb_GetMdfMode(&mdfMode);
        if (mdfMode == DBR_MDF_DISABLE_MODE)
        {
            ti_igmp_start();
        }

        if (DBR_REGISTERED_MODE != dbridge_db.mode)
        {
            struct net_device * dev = dev_get_by_name (CABLE_DEV_NAME);
            ti_hil_pp_event (TI_BRIDGE_PORT_FORWARD, (void *)dev);
            if(dev)
                dev_put(dev);
        }
    }
    else
    {
        ti_igmp_stop();
        if (DBR_REGISTERED_MODE == dbridge_db.mode)
        {
            struct net_device * dev = dev_get_by_name (CABLE_DEV_NAME);
            ti_hil_pp_event (TI_BRIDGE_PORT_DISABLED, (void *)dev);
            if(dev)
                dev_put(dev);
        }
        if(selective_packet_handler)
        {
            ti_unregister_selective_packet_handler (LBRIGE_CONNECTION_DEV_NAME, ti_selective_forward, 1);
            selective_packet_handler = 0;
        }
        if(register_docsis_egress_hook)
        {
            configure_docsis_egress_hook_handler(0);
            register_docsis_egress_hook =0;
        }

    }

    dbridge_db.mode=mode;

}

/*************************************************************************************/
/*! \fn iint DbridgeDb_GetMode (void)
 *************************************************************************************
 *  \brief get docsis bridge mode
 *  \param[in] none
 *  \return docsis bridge mode - off, registered pre_registration or naco_off.
 *************************************************************************************/
int DbridgeDb_GetMode(void)
{
    return dbridge_db.mode;
}

/*************************************************************************************/
/*! \fn iDbridgeCounter_t* Dbridge_GetCounters(void)
 *************************************************************************************
 *  \brief get docsis bridge counters
 *  \param[in] none
 *  \return pointer to the Dbridge counter structure.
 *************************************************************************************/
DbridgeCounters_t* Dbridge_GetCounters(void)
{
    return &dbridge_counters;
}

/***************************************************************************/
/*! \fn int DbridgeDB_GetlbridgeInterfaceMask(void)                                     
 **************************************************************************
 *  \brief Get lbridge Interface Mask.
 *  \param[in] None .
 *  \param[out] None.
 *  \return lbridge Interface Masks.
 **************************************************************************/
int DbridgeDB_GetlbridgeInterfaceMask(void)
{
    return lbridge_interface_mask;  
}

/***************************************************************************/
/*! \fn void DbridgeDb_SetMode (int mode)
 ***************************************************************************
 *  \brief set docsis bridge erouter mode
 *  \param[in] docsis bridge erouter mode - enable/disable 
 *  \return none.
 ***************************************************************************/
void DbridgeDb_Erouter_SetMode(int mode)
{
    dbridge_db.eroter_mode = mode;
}

/***************************************************************************/
/*! \fn iint DbridgeDb_Erouter_GetMode (void)
 ***************************************************************************
 *  \brief get docsis bridge erouter mode
 *  \param[in] none
 *  \return docsis bridge erouter mode enable/disable.
 ***************************************************************************/
int DbridgeDb_Erouter_GetMode(void)
{
        return dbridge_db.eroter_mode;
}

#ifdef CONFIG_TI_DOCSIS_EGRESS_HOOK

static int configure_docsis_egress_hook_handler(int reg)
{

	/* Install the Ingress Hook. */
	{
		extern int ti_register_docsis_egress_hook_handler (struct net_device* dev, int (*docsis_egress_hook)(struct sk_buff *skb));
        extern int ti_deregister_docsis_egress_hook_handler (struct net_device* dev);

        if(NULL == eth0_netdev)
        {
            eth0_netdev= dev_get_by_name (ETH_DEV_NAME);
        }
		if(eth0_netdev)
		{
            if(reg)
            {
                if (ti_register_docsis_egress_hook_handler (eth0_netdev, Dbridge_egress_hook) == 0)
                    printk ("register ingress Hook on  %s \n",eth0_netdev->name);
            }
            else
            {
                if (ti_deregister_docsis_egress_hook_handler (eth0_netdev) == 0)
                    printk ("deregister ingress Hook on  %s \n",eth0_netdev->name);
                dev_put(eth0_netdev);
                eth0_netdev = NULL;
            }
		}
        else
        {
            printk ("warning: configure ingress hook fail no eth dev %s \n",usb0_netdev->name);
        }

        if(NULL == usb0_netdev)
        {
            usb0_netdev= dev_get_by_name (USB_DEV_NAME);
        }
		if(usb0_netdev)
		{
            if(reg)
            {
                if (ti_register_docsis_egress_hook_handler (usb0_netdev, Dbridge_egress_hook) == 0)
                    printk ("register ingress Hook on  %s \n",usb0_netdev->name);
            }
            else
            {
                if (ti_deregister_docsis_egress_hook_handler (usb0_netdev) == 0)
                    printk ("deregister ingress Hook on  %s \n",usb0_netdev->name);
                dev_put(usb0_netdev);
                usb0_netdev = NULL;
            }
		}
        else
        {
            printk ("warning: configure ingress hook fail no usb dev %s \n",usb0_netdev->name);
        }

	}
    return 0;

}

/**************************************************************************/
/*! \fn int Dbridge_egress_hook(struct sk_buff *skb)
 **************************************************************************
 *  \brief   In
 *  \param[in] SKB.
 *  \return 
 **************************************************************************/
static int Dbridge_egress_hook(struct sk_buff *skb)
{
    struct ethhdr* ptr_ethhdr;

    if(skb)
    {
        ptr_ethhdr = (struct ethhdr *)skb->mac.raw;

        if((skb->input_dev == eth0_netdev) || (skb->input_dev == usb0_netdev))
        {   
            dbridge_counters.DevCounters[skb->input_dev->ifindex].In++;
            dbridge_counters.DevCounters[skb->dev->ifindex].Out++;
            if ( !MAC_ISMULTICAST(ptr_ethhdr, h_dest) )
            {/*If this is unicast packet - increment unicast counters as well*/
                dbridge_counters.DevCounters[skb->input_dev->ifindex].inUcast++;
                dbridge_counters.DevCounters[skb->dev->ifindex].outUcast++;
            } else
            {  /*If this is broadcast packet - increment broadcast counter as well*/
                if ( MAC_ISBROADCAST(ptr_ethhdr, h_dest) )
                {
                    dbridge_counters.DevCounters[skb->input_dev->ifindex].inBcast++;
                    dbridge_counters.DevCounters[skb->dev->ifindex].outBcast++;
                } else
                {   /*If this is multicast packet - increment multicast counter as well*/
                    dbridge_counters.DevCounters[skb->input_dev->ifindex].inMcast++;
                    dbridge_counters.DevCounters[skb->dev->ifindex].outMcast++;
                }
            }
        }
    }
    return 0;
}

/**************************************************************************/
/*! \fn void DbrigdeDb_AltDelEntry(DbridgeAltEntry_t *altEntry)
 **************************************************************************
 *  \brief Remove ALT entry from list
 *  \param[in] altEntry - entry to be removed
 *  \return void
 **************************************************************************/
static void DbrigdeDb_AltDelEntry(DbridgeAltEntry_t *altEntry)
{
    unsigned char cpe_num;

    /* Decrease CPE counters */
    if (altEntry->macLearningFrom==CPE_DYNAMIC)
        dbridge_counters.DynamicCpe--;
    else if (altEntry->macLearningFrom==CPE_STATIC)
        dbridge_counters.StaticCpe--;

    /* Remove from OCCUPIED list */
    list_del(&altEntry->listAltEntry);

    /* Preserve cpe num */
    cpe_num = altEntry->cpe_number;

    /* 0 */
    memset (altEntry, 0, sizeof(*altEntry));

    /* Restore cpe num */
    altEntry->cpe_number = cpe_num;

    /* Add to FREE list */
    /* add_tail, so reuse will be delayed till all other entries are used */
    list_add_tail(&altEntry->listAltEntry, &DBRALT_FREE_HEAD);
}



#endif /* CONFIG_TI_DOCSIS_EGRESS_HOOK */

EXPORT_SYMBOL(DbridgeDb_DevGetByInterface);
