/*
 *
 * hal_interrupt_cause.h
 * Description:
 * declaration of functions and types used in the interrupt cause-detection module
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#ifndef _HAL_INTERRUPT_CAUSE_H_
#define _HAL_INTERRUPT_CAUSE_H_

#include "_tistdtypes.h"

/**************************************************************************/
/*      INTERFACE  Defines and Structs                                    */
/**************************************************************************/

#define MAC_ERROR_INT_BUNDLE_NUMBER 4


typedef unsigned int IntCauseErrorGroup_t[MAC_ERROR_INT_BUNDLE_NUMBER];
 /* each field will hold the bitmap of the cause-interrupts for one of the bundles*/

/* A structure that defines receiver parameters used by the SW. */
typedef struct RecieverFecParam {
	Bool value;
	Uint32 intr;
	Uint32 lockbit;
	Uint32 lostbit;
} RecieverFecParam_t;


/* A structure that defines the interrupt Cause and corresponding Mask register . */
/* For PHY Upstream (TX) interrupts.                                              */
typedef struct PhyTxMask {
	Uint32 mask;
	Uint32 cause;
} PhyTxMask_t;

/* Macro for retrieving the global index of an interrupt cause.         */
/* The macro receives the interrupt-group and the specific bit-number   */
/* in that group, and returns the "global" index.                       */
#define HAL_GET_INTERRUPT_CAUSE(group, bitNumber)   (((group)*32)+(bitNumber))

/* Macro for retrieving the interrupt-group of an interrupt.            */
/* The macro receives the global interrupt-cause and returns the group  */
/* index.                                                               */
#define HAL_GET_INTERRUPT_GROUP(cause)              ((cause)/32)

/* Macro for retrieving the bit number of an interrupt inside its group.*/
/* The macro receives the global interrupt-cause and returns the local  */
/* bit-index.                                                           */
#define HAL_GET_INTERRUPT_BIT(cause)                ((cause)%32)

/* Macro for retrieving the index number of an interrupt inside database*/
/* of type PhyFecLockLostStatus_t. The macro receives phy-interrupt     */
/* number and returns the local db-index.                               */
#define HAL_PHY_GET_FEC_STATUS_DB_INDX(intrBit)     ((intrBit)/6)

/* Macro for retrieving the status of phy interrupt inside database     */
/* of type PhyFecLockLostStatus_t. It indicates if the interrupt has of */
/* type FEC-Lock or FEC-Lost. The macro receives phy-interrupt number.  */
#define HAL_PHY_IS_FEC_LOCK(intrBit)                ((intrBit)%6)


#define MAC_ERR1_INT_GROUP                          0
#define MAC_ERR2_INT_GROUP                          1
#define MAC_ERR3_INT_GROUP                          2
#define MAC_ERR4_INT_GROUP                          3
#define MAC_UCD_INT_GROUP                           4
#define MAC_US_FW_INT_GROUP                         5
#define MAC_DS_FW_INT_GROUP                         6
#define PHY_INT_GROUP                               7
#define PHY_PGA_GRT_INT_GROUP                       8
#define REMOTE_PHY_INT_GROUP                        9
#define PHY_MPT_INT_GROUP                           10
#define REMOTE_MAC_INT_GROUP                        11
                                                    
/* TODO: should this be defined here? */
/* MAPPING of the DOCSIS Interrupts to IRQ lines at the SoC Interrupt Controller */
/* From the Puma5 Device Specification ("Chip-Global" chapter):                  */
/* 15	DMACERRINT	Active high level	Docsis MAC	docsis_mac_hst_error_int     */
/* 16	DMACUCDINT	Active high level	Docsis MAC	docsis_mac_hst_new_ucd_int   */
/* 17	DMACARMINT	Active high level	Docsis MAC	docsis_arm_interrupt         */
/* 18	DMAPDSPINT	Active high level	Docsis MAC	docsis_pdsp_interrupt        */
/* 19	DPHYINT	    Active high level	Docsis PHY	docsis_phy_interrupt         */
/* 43	GRTINT	    Active falling Edge	Docsis SOC	docsis_pga_grt_interrupt     */
#define MAC_ERR_INT_IRQ_NUM                         15
#define MAC_UCD_INT_IRQ_NUM                         16
#define MAC_US_FW_INT_IRQ_NUM                       17
#define MAC_DS_FW_INT_IRQ_NUM                       18
#define PHY_INT_IRQ_NUM                             19
#define VLYNQ_INT_IRQ_NUM                           20
#define PHY_PGA_GRT_INT_NUM                         (pgaGrtIntNum)
#define PHY_PGA_GRT_INT_NUM_REV1                    1
#define PHY_PGA_GRT_INT_NUM_REV2                    43
#define PHY_MPT_INT_NUM                             1

/* For the GRT interrupt - there is only one interrupt related to this IRQ, */
/* therefore the status mask will always return that bit 0 is active.       */
#define PHY_PGA_GRT_INT_MASK                       0x1
#define PHY_PGA_GRT_INT_INDEX                      0

/* If working with system REV-1, the interrupt is muxed with GPIO */
#define PHY_PGA_GRT_INT_REV1_GPIO                  27

/* For the MPT interrupt - there is only one interrupt related to this IRQ, */
/* therefore the status mask will always return that bit 0 is active.       */
#define PHY_MPT_INT_MASK                           0x1
#define PHY_MPT_INT_INDEX                          0

/**************************************************************************/
/*      INTERFACE VARIABLES (prefix with EXTERN)                          */
/**************************************************************************/


/**************************************************************************/
/*      INTERFACE FUNCTIONS Prototypes:                                   */
/**************************************************************************/

/**********************************/               
/*  MAC Errors Interrupt Group    */               
/**********************************/               
/*! \fn Int32 HAL_IsrGetCauseMacErrorGroup(IntCauseErrorGroup_t *pCause)
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC-error group interrupt-line.
 *  \param[out] pCause pointer to the structure that will hold the interrupt cause(s).
 *  \return OK if successful, reject or fail otherwise                                     
 */
Int32 HAL_IsrGetCauseMacErrorGroup(IntCauseErrorGroup_t *pCause);


/****************************************/               
/*    MAC UCD-Change Interrupt group    */               
/****************************************/               
/*! \fn Uint32 HAL_IsrGetCauseUcdChange()
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC UCD Change interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 */
Uint32 HAL_IsrGetCauseUcdChange( void );


/***********************************/               
/*    MAC US FW Interrupt group    */               
/***********************************/               
/*! \fn Uint32 HAL_IsrGetCauseUsFw()
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC US FW interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 */
Uint32 HAL_IsrGetCauseUsFw( void );


/***********************************/               
/*    MAC DS FW Interrupt group    */               
/***********************************/               
/*! \fn Uint32 HAL_IsrGetCauseDsFw()
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \MAC DS FW interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 */
Uint32 HAL_IsrGetCauseDsFw( void );


/***********************************/               
/*         PHY  Interrupts         */               
/***********************************/               
/*! \fn Uint32 HAL_IsrGetCausePhy(Bool isFromRemote)
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \PHY interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 */
Uint32 HAL_IsrGetCausePhy( Bool isFromRemote );


/*! \fn Uint32 HAL_IsrGetCauseRemoteMac()
 *  \brief retreives the specific causes that asserted an interrupt on the
 *  \Remote MAC interrupt-line.
 *  \return The interrupt cause(s) as a bit-map.
 */
Uint32 HAL_IsrGetCauseRemoteMac( void );


/***********************************/               
/*      Mask/Unmask Service        */               
/***********************************/               
/*! \fn Int32 HAL_IsrMaskInterrupt(Uint32 cause, Bool maskOp)
 *  \brief Masks or unmasks a speific interrupt.
 *  \param[in] cause defines the specific interrupt that needs to be 
 *  \                masked/unmasked.
 *  \param[in] maskOp defines whether to mask or unmask. 
 *  \                 True = mask, False= unmask. 
 *  \return OK if successful, reject or fail otherwise                                     
 */
Int32 HAL_IsrMaskInterrupt(Uint32 cause, Bool maskOp);


/**************************************************************************/
/*! \fn int HAL_IsrResetPhyFecDB()
 **************************************************************************
 *  \brief Reset Phy-Rx database.
 *  \param[in] none.
 *  \return 0 if successful, reject or fail otherwise.
 **************************************************************************/
Int32 HAL_IsrResetPhyFecDB(void);
#endif
