/***************************************************************************
**+----------------------------------------------------------------------+**
**|                                ****                                  |**
**|                                ****                                  |**
**|                                ******o***                            |**
**|                          ********_///_****                           |**
**|                           ***** /_//_/ ****                          |**
**|                            ** ** (__/ ****                           |**
**|                                *********                             |**
**|                                 ****                                 |**
**|                                  ***                                 |**
**|                                                                      |**
**|     Copyright (c) 2006-2009 Texas Instruments Incorporated           |**
**|                        ALL RIGHTS RESERVED                           |**
**|                                                                      |**
**| Permission is hereby granted to licensees of Texas Instruments       |**
**| Incorporated (TI) products to use this computer program for the sole |**
**| purpose of implementing a licensee product based on TI products.     |**
**| No other rights to reproduce, use, or disseminate this computer      |**
**| program, whether in part or in whole, are granted.                   |**
**|                                                                      |**
**| TI makes no representation or warranties with respect to the         |**
**| performance of this computer program, and specifically disclaims     |**
**| any responsibility for any damages, special or consequential,        |**
**| connected with the use of this program.                              |**
**|                                                                      |**
**+----------------------------------------------------------------------+**
***************************************************************************/

/**************************************************************************/
/*      INCLUDES                                                          */
/**************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/select.h>
#include <string.h>

#include "icc.h"
#include "context_owners.h"
#include "hal_global.h"
#include "init_process.h"
#include "docsis_logger.h"
#include "logger.h"
#include "docsis_shared_dbs.h"
#include "pcdapi.h"
#include "sys_types.h"
#include "sys_cfg_db.h"
#include "downstream_manager_api.h"
#include "hal_tuner_mgr_demod_api.h"
#include "mask_lib.h"
#include "hal_interrupt_cause.h"
#include "hal_phy_dnstream.h"
#include "chdb_main.h"
/**************************************************************************/
/*      DEFINES:                                                          */
/**************************************************************************/
#define MXL24X_FEC_LOST_BIT  0
#define MXL24X_FEC_LOCK_BIT  1

/**************************************************************************/
/*      IMPLEMENTATION                                                    */
/**************************************************************************/
void HAL_IsrHandleMptInterrupt(ICC_context iccContext)
{
    Uint32 receiverId, maxReceivers;
    Bool isLocked;
    Bool isMpegTuner = False;
    ICC_message_t *msg;

    SysCfgDb_GetBoardConfigNumOfReceivers(&maxReceivers);

    while (1)
    {
        if (ICC_wait_message(iccContext, &msg, ICC_TIMEOUT_FOREVER) == -1)
        {
            TI_LOG_ERROR(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0, "HAL_IsrHandleMptInterrupt error receive message");
        }
        else
        {
            ICC_dispose_message(msg);

            TI_LOG_INFO(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0, "MPT-ISR :: HAL_IsrHandleMptInterrupt receive message");

            /* go over all receivers and clear MPEG-Tuner interrupts */
            for(receiverId = 1; receiverId <= maxReceivers; receiverId++)
            {
                SysCfgDb_GetBoardConfigIsMpegTunerReceiver(receiverId,&isMpegTuner);

                /* If the receiver is an MPEG-Tuner and "owned" by DOCSIS */
                if( isMpegTuner )
                {
                    /* Clear FEC-Loss and FEC-Lock interrupts */
                    HAL_TunerMgrDemodClearIntr(receiverId-1, MXL24X_FEC_LOST_BIT);
                    HAL_TunerMgrDemodClearIntr(receiverId-1, MXL24X_FEC_LOCK_BIT);
                }
            }
            
            /* clear the MPT interrupt and unmask it */
            HAL_IsrClearIrq(PHY_MPT_INT_NUM);
            HAL_IsrMaskCause((HAL_GET_INTERRUPT_CAUSE(PHY_MPT_INT_GROUP, PHY_MPT_INT_INDEX)), False);
            
            /* Get FEC status and generate events */
            for(receiverId = 1; receiverId <= maxReceivers; receiverId++)
            {
                SysCfgDb_GetBoardConfigIsMpegTunerReceiver(receiverId,&isMpegTuner);
                
                if( isMpegTuner )
                {
                    HAL_TunerMgrDemodGetFecLock(receiverId-1, &isLocked);
                    
                    if(isLocked)
                    {
                        TI_LOG_INFO(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0,"MPT-ISR :: QamFecRecovery portId %d.", receiverId);
                        DSM_NotifyQamFecRecovery(receiverId);
                    }
                    else
                    {
                        TI_LOG_INFO(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0,"MPT-ISR :: QamFecLost portId %d.", receiverId);
                        DSM_NotifyQamFecLost(receiverId);
                    }
                }
            }
        }
    }
}




int main(int argc, char **argv)
{
    ICC_context idp;

    if( PCD_api_find_process_id(argv[0]) )
    {
        fprintf( stderr, "Mpeg-Tuner interrupt handler already active");
        return 1;
    }

    PCD_api_register_exception_handlers( argv[0], NULL );


    if (ICC_init() < 0) 
    {
		printf("failed to init ICC, errno %d\n", errno);
		exit(-4);
	}

    if (ICC_create_destination_point(&idp) == -1)
    {
        TI_LOG_FATAL(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0, "ICC_create_destination_point for MPT failed");
        exit(-5);
    }
    if (ICC_set_owner(idp, CTX_MPT) == -1) 
    {
        TI_LOG_FATAL(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0, "ICC_set_owner for MPT failed");
        exit(-6);
    }

    INIT_PROCESS(INIT_ALL_ITEMS);

    if( HAL_GlobalRegistration(True) != STATUS_OK )
    {
        printf("Global registration failed\n");
        exit(-7);
    }

    if(DocsisCfgDb_RetrieveAccess() == STATUS_NOK)
    {
        TI_LOG_ERROR(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0, "Failed to retrieve access to DOCSIS CFG DB");
        return STATUS_NOK;
    }

    if(CHDB_Main_RetrieveAccess() == STATUS_NOK)
    {
        TI_LOG_ERROR(TI_COMPONENT_DOCSIS, DOCSIS_MOD_HAL, 0, "Failed to retrieve access to CHDB");
        return STATUS_NOK;
    }

    HAL_IsrHandleMptInterrupt(idp);

    return 0;
}



