/*
 *
 * docsis_mng.c
 * Description:
 * DOCSIS management driver implementation
 *
 * Copyright (C) 2008 Texas Instruments Incorporated - http://www.ti.com/ 
 * 
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
*/

#include <asm/uaccess.h>
#include <linux/kernel.h>
#include <linux/init.h> 
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/poll.h>
#include <linux/interrupt.h>
#include <linux/cdev.h>
#include <linux/sched.h>
#include <asm/io.h>
#include <linux/timer.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>

#include <puma5/puma5_cppi.h>
#include "_tistdtypes.h"
#include "mng_cppi.h"
#include "pal.h"
#include "pal_cppi41.h"
#include "pal_osWait.h"
#include "docsis_mng.h"
#include "mac_mngt_msg.h"

#define DOCSIS_MNG_MAJOR 32
#define DOCSIS_MNG_NAME "docsis_mng"

/*
 * use new defines from puma5_cppi_prv.h
 */
#define DOCSIS_MNG_RX_BD_SIZE           CMGR_CPPI4x_RX_HOST_BD_SIZE /* sizeof Cppi4HostDesc + align */
#define DOCSIS_MNG_RX_BD_NUM            64 // cable_pp: there is no region with 32 bds, DMAC_MNG_RX_HOST_BD_NUM /* can we have less than that ??? */

#define DOCSIS_MNG_TX_BD_SIZE           CMGR_CPPI4x_TX_HOST_BD_SIZE /* sizeof Cppi4HostDesc + align */
#define DOCSIS_MNG_TX_BD_NUM            64 // cable_pp: there is no region with 32 bds, DMAC_MNG_TX_HOST_BD_NUM /* can we have less than that ??? */

#define DOCSIS_MNG_ACCRX_INTD_VEC       5 //CMNG_ACC_RX_INTV
#define DOCSIS_MNG_ACC_RX_CHNUM         9 //CMNG_ACC_RX_CHNUM

#define DOCSIS_MNG_RX_BUFFER_SIZE       2*1024 /* max msg size ???*/
#define DOCSIS_MNG_RX_BUFFER_NUM        DOCSIS_MNG_RX_BD_NUM

#define DOCSIS_MNG_TX_BUFFER_SIZE       2*1024 /* max msg size ???*/
#define DOCSIS_MNG_TX_BUFFER_NUM        DOCSIS_MNG_TX_BD_NUM

/* Only CPPI specified bytes need to be invalidated */
#define CPPI4_BD_LENGTH_FOR_CACHE       32


/* accumulator defines */
#define CPPI41_ACC_ENTRY_D              0 /* should be in the pal !!!*/
#define CPPI41_ACC_ENTRY_CD             1
#define CPPI41_ACC_ENTRY_ABCD           2

#define DOCSIS_MNG_ACC_ENTRY_TYPE   CPPI41_ACC_ENTRY_D
#define DOCSIS_MNG_ACC_MAXPAGENTRY  4
#define DOCSIS_MNG_ACC_ENTRY_SIZE   4 /* 4 bytes pointer */

#define DOCSIS_MNG_ACC_INTR_MODE    2
#define DOCSIS_MNG_ACC_INTR_DELAY   0
#define DOCSIS_MNG_ACC_STALL_AVOID  0
#define DOCSIS_MNG_ACC_TERM_MODE    0
#define DOCSIS_MNG_ACC_PAGE_COUNT   2

#define DOCSIS_MNG_RX_INTR_LINE             (AVALANCHE_INTD_BASE_INT + DOCSIS_MNG_ACCRX_INTD_VEC)

/* Protocol specific bits - US */
#define US_PSI_PRIVACY_OFF_BIT_MASK        0x00008000 /* bit 15 */
#define US_PSI_SFINDEX_MASK                0xFF000000 /* bits 31-24 */
#define US_PSI_SFINDEX_SHIFT               24

/* Protocol specific bits - DS */
#define DS_PROTOCOL_SPECIFIC_TS_PRESENT    0x00000020 /* bit 5 */
#define DS_PROTOCOL_SPECIFIC_TS_SIZE       4          /* Timestamp size in bytes */
#define DS_PROTOCOL_SPECIFIC_PORT_SHIFT    8          /* Port is located in bits 8..15 */
#define DS_PROTOCOL_SPECIFIC_PORT_SIZE     1          /* Port size in bytes */

DECLARE_WAIT_QUEUE_HEAD(docsis_mng_wait);
static int ref = 0;
static struct semaphore wr_sem;

/**
 *  Host descriptor structure. 
 * Note: This structure can represent either Host Packet Descriptor or Host 
 * Buffer Descriptor at a time, so validity of some fields depends on the 
 * type of descriptor.
 */
typedef volatile struct {
    /* hardware descriptor fields from this point */
    Uint32 descInfo;     /**< Desc type, proto specific word cnt, pkt len (valid only in Host PD)*/
    Uint32 tagInfo;      /**< Source tag (31:16), Dest Tag (15:0) (valid only in Host PD)*/
    Uint32 pktInfo;      /**< pkt err state, type, proto flags, return info, desc location */
    Uint32 buffLen;      /**< Number of valid data bytes in the buffer */
    Uint32 bufPtr;       /**< Pointer to the buffer associated with this descriptor */
    Ptr nextBDPtr;       /**< Pointer to the next buffer descriptor */
    Uint32 orgBuffLen;   /**< Original buffer size */
    Uint32 orgBufPtr;    /**< Original buffer pointer */
    Uint32 netInfoWord0; /**< Network stack private communications info (valid only in Host PD) */
    Uint32 netInfoWord1; /**< Network stack private communications info (valid only in Host PD) */
    
    /* Protocol specific data */
    Uint32 protocolSpec0;
    Uint32 protocolSpec1;
    Uint32 protocolSpec2;
    
    /* software only fields from this point */

} DocsisCppi4HostDesc;

static PAL_Handle palHandle;
static PAL_Handle palHandleDocsis;

static Cppi4Queue           rxfdQueue = {CNI_MGMT_CPPI4x_FBD_QMGR,   CNI_MGMT_CPPI4x_FBD_QNUM(0)};
static Cppi4Queue           rxQueue =   {CNI_MGMT_CPPI4x_RX_QMGR,    CNI_MGMT_CPPI4x_RX_QNUM(0) };
static PAL_Cppi4QueueHnd    rxfdQueueHdl;
static PAL_Cppi4QueueHnd    rxQueueHdl;

static Cppi4Queue           txfdQueue = {CNI_MGMT_CPPI4x_FBD_QMGR,   CNI_MGMT_CPPI4x_FBD_QNUM(1)};
static Cppi4Queue           txQueue =   {CNI_MGMT_CPPI4x_TX_QMGR,    CNI_MGMT_CPPI4x_TX_QNUM(0) };
static PAL_Cppi4QueueHnd    txfdQueueHdl;
static PAL_Cppi4QueueHnd    txQueueHdl;

static Ptr txbdAreaAddrs;
static Ptr txbufferAreaAddrs;

/* this is the place to set data offset */
/* Ilan <- consider to make it configurable parameter !!! */
static int txBufferOffset = 0; 

static DocsisCppi4HostDesc* curBD4Read = NULL;
static Bool dataReady = False;

int debug_enable = 0;

#ifdef POOLING
struct timer_list mngPoolTimer;
#endif

#ifdef _USE_ACC_LIST
static Uint8 accWrkArea[DOCSIS_MNG_ACC_MAXPAGENTRY*DOCSIS_MNG_ACC_PAGE_COUNT*4];
static Uint32 accCurPage = 0;
static Uint32 accCurBdOnPage = 0;
static Uint32 accDataListPtr[DOCSIS_MNG_ACC_PAGE_COUNT];
static Cppi4Queue  accRxQueue = {DOCSIS_MNG_QMGR,DOCSIS_MNG_RX_QNUM1};
static PAL_Cppi4QueueHnd accRxQueueHdl;
#endif

static DocsisCppi4HostDesc* docsisMngGetNextBd(void);
static Int32 HAL_DocsisCppiInit(void);

static int (* HAL_DocsisQosClass_ClassifyMacMngCB)(Uint8 *buf,Int16 *sfIndex,Int16 *phsIndex) = NULL;

irqreturn_t
docsisMngIsr(int irq, void *dev_id, struct pt_regs *regs)
{
    /* wake up (unblock) for reading data from userspace   */
    dataReady = True;
    wake_up_interruptible(&docsis_mng_wait);
    return IRQ_HANDLED;
}

/*
 * ref count the total number of opens (both read and write).
 * It makes the code simpler, by allocating all the needs of r/w
 * at open time and dealocating them (together) at the last close.
 */
static int
docsisMngOpen(struct inode *inode, struct file *file) 
{
    DocsisCppi4HostDesc *currBD = NULL;
    volatile static UINT32 bufferCount = 0;
    int r = 0;

    down(&wr_sem);
    printk(KERN_INFO "docsisMngOpen: ref %d\n", ref);
    if (++ref > 1) {
        up(&wr_sem);
        return r;
    }

    currBD = (DocsisCppi4HostDesc *)PAL_cppi4QueuePop(rxfdQueueHdl);
    if (currBD != NULL) {
        currBD->bufPtr = currBD->orgBufPtr;
        currBD->buffLen = 20;
        *((UINT32 *)currBD->bufPtr) = bufferCount;
        bufferCount++;
        PAL_cppi4QueuePush(rxQueueHdl, (Ptr)PAL_CPPI4_VIRT_2_PHYS((Uint32)currBD), 
            DOCSIS_MNG_QM_HDESC_SIZE_VAL, 0);
    }

#ifdef POOLING
    mngPoolTimer.expires = jiffies + HZ / 100;
    add_timer(&mngPoolTimer);
    printk(KERN_INFO "docsisMngOpen: start pool timer\n");
#else
    if (r = avalanche_intd_enable_interrupt(0, DOCSIS_MNG_ACC_RX_CHNUM) {
        printk(KERN_WARNING "docsisMngOpen: Unable to enable INTD interrupt \n");
        up(&wr_sem);
        return r;
    }

    /* register to the interrupt on the apdsp !!!*/
    if (r = request_irq(LNXINTNUM(DOCSIS_MNG_RX_INTR_LINE), docsisMngIsr, SA_INTERRUPT, "docsis mng", NULL)) {
        printk(KERN_WARNING "Failed to register the irq for docsis mng.\n");
        up(&wr_sem);
        return r;
    }
    printk(KERN_INFO "docsisMngOpen: set intr on line : %d\n , %d\n", LNXINTNUM(DOCSIS_MNG_RX_INTR_LINE),DOCSIS_MNG_RX_INTR_LINE);
#endif

    up(&wr_sem);
    return 0;
}

/* close function - called when the "file" /dev/docsis_mng is closed in userspace   */
static int
docsisMngRelease(struct inode *inode, struct file *file) 
{
    down(&wr_sem);
    printk(KERN_INFO "docsisMngRelease: ref %d\n", ref);
    if (--ref) 
    {
        up(&wr_sem);
        return 0;
    }
    up(&wr_sem);
    return 0;
}

static
DocsisCppi4HostDesc* docsisMngGetNextBd()
{
#ifndef _USE_ACC_LIST
    DocsisCppi4HostDesc *curBD = (DocsisCppi4HostDesc *)PAL_cppi4QueuePop(rxQueueHdl);
    if (curBD == NULL)
    {
        dataReady = False;
    }
    else
    {
        curBD = PAL_CPPI4_PHYS_2_VIRT(curBD);
    }
    return(curBD);
#else
    /* all acc handling should be moved to the pal !!! */
    /* I don't want to duplicate it here    but i NEED to :( */
    Uint32 curBD;

    curBD = accDataListPtr[accCurPage] + accCurBdOnPage*DOCSIS_MNG_ACC_ENTRY_SIZE;
    if (debug_enable) {
        printk(KERN_INFO "docsisMngGetNextBd: curBd =%x \n",curBD);
        printk(KERN_INFO "docsisMngGetNextBd: accCurPage =%x, accCurBdOnPage=%x \n",accCurPage,accCurBdOnPage);
    }
    if (((Ptr)curBD) != NULL) {
        curBD = PAL_CPPI4_PHYS_2_VIRT(curBD);
        accCurBdOnPage++;
        return((DocsisCppi4HostDesc *)(curBD & QMGR_QUEUE_N_REG_D_DESC_ADDR_MASK));
    }
    /* last entry on the page */
    /* invalidate the working area page */
    PAL_CPPI4_CACHE_INVALIDATE(accDataListPtr[accCurPage], DOCSIS_MNG_ACC_MAXPAGENTRY * 4);

    accCurPage++;
    if (accCurPage>=DOCSIS_MNG_ACC_PAGE_COUNT)
        accCurPage = 0;
    accCurBdOnPage = 0;
    dataReady = False;
    /* decrement the interrupt counter page */
    avalanche_intd_set_interrupt_count (0, DOCSIS_MNG_ACC_RX_CHNUM, 1);
    /* clear the interrupt mask to get next interrupt */
    avalanche_intd_write_eoi (DOCSIS_MNG_ACC_RX_CHNUM);
    return(NULL);
#endif
}

/* read function called when from /dev/docsis_mng is read */
static ssize_t
docsisMngRead(struct file *file, char *buf, size_t count, loff_t *ppos) 
{
    static int  offset = 0; 
    Uint32      user_count = count; /* save original count specified by the user */
    Uint32      timestamp;          /* timestamp associated with the packet */
    Uint8       portId;             /* Port where message arrived */
    int         err;
    
    /* i can add a semaphore to protect the critical section, but should I ??? */
    /* is it possible to have two entities reading from this device ??? */

    if (debug_enable)
        printk(KERN_INFO "docsisMngRead: in with buffer size %d\n",count);
    /* if we are not in the middle of the message , go bring a new one*/
    if (curBD4Read == NULL) {
        /* get the message from queue */
        while ((curBD4Read = docsisMngGetNextBd()) == NULL) {
            if (file->f_flags & O_NONBLOCK)
                return(-EAGAIN);

            /* do i have an issue with race condition ??? dataReady*/
            /* need to change to wait_event_interruptible(docsis_mng_wait, condition); */
            if (debug_enable)
                printk(KERN_INFO "docsisMngRead: go sleep %x\n",(Uint32)curBD4Read);            
            wait_event_interruptible(docsis_mng_wait, (dataReady == True));
            if (debug_enable)
                printk(KERN_INFO "docsisMngRead: i am wake !\n");           
        }
        offset = 0;
    }
    /* invalidate the cache bd */
    PAL_CPPI4_CACHE_INVALIDATE(curBD4Read, CPPI4_BD_LENGTH_FOR_CACHE);
    if (debug_enable)
        printk(KERN_INFO "docsisMngRead: processing %x\n, data %x,size %d",
            (Uint32)curBD4Read, (Uint32)(PAL_CPPI4_PHYS_2_VIRT(curBD4Read->bufPtr)),
            curBD4Read->buffLen);           

    /* message is smaller than buffer */
    if (count > (curBD4Read->buffLen - offset))
        count = (curBD4Read->buffLen - offset); 

    PAL_CPPI4_CACHE_INVALIDATE(PAL_CPPI4_PHYS_2_VIRT(curBD4Read->bufPtr), curBD4Read->buffLen);
	
    /* should i change the addr from phy to virt??? probably yes */
    err = copy_to_user(buf, (char*)PAL_CPPI4_PHYS_2_VIRT(curBD4Read->bufPtr) + offset, count);
    if (debug_enable)
        printk(KERN_INFO "docsisMngRead: copy to user %d bytes with offset %x\n",count,offset);         
    
    /* lets keep the message for the next read ???*/
    if (err != 0)
        return -EFAULT;

    /* if we done with the message */
    if (count == (curBD4Read->buffLen - offset)) {
        /* check if timestamp present */
        if (curBD4Read->protocolSpec0 & DS_PROTOCOL_SPECIFIC_TS_PRESENT) {
            if ((count + DS_PROTOCOL_SPECIFIC_TS_SIZE)> user_count) {
                if (debug_enable)
                    printk(KERN_WARNING "docsisMngRead: Not enough space to copy timestamp to user!\n");
            }
            else {
                /* get timestamp from BD and append it at end of packet */
                timestamp = curBD4Read->protocolSpec1;
                err = copy_to_user(buf + count, (char*)&timestamp, DS_PROTOCOL_SPECIFIC_TS_SIZE);
                if (debug_enable)
                    printk(KERN_INFO "docsisMngRead: Copy timestamp 0x%x to user (%d bytes) - %s.\n",
                                            timestamp, DS_PROTOCOL_SPECIFIC_TS_SIZE, err ? "failed" : "ok");         
                if (err == 0)
                    count += DS_PROTOCOL_SPECIFIC_TS_SIZE;
            }
        }

        if ((count + DS_PROTOCOL_SPECIFIC_PORT_SIZE)> user_count) {
            if (debug_enable)
                printk(KERN_WARNING "docsisMngRead: Not enough space to copy port to user!\n");
        }
        else {
            portId = (curBD4Read->protocolSpec0 >> DS_PROTOCOL_SPECIFIC_PORT_SHIFT) + 1;
            err = copy_to_user(buf + count, (char*)&portId, DS_PROTOCOL_SPECIFIC_PORT_SIZE);
            if (debug_enable)
                printk(KERN_INFO "docsisMngRead: Copy port %d to user (%d bytes) at offset %d - %s.\n",
                                         portId, DS_PROTOCOL_SPECIFIC_PORT_SIZE, count, err ? "failed" : "ok");         
            if (err == 0)
                count += DS_PROTOCOL_SPECIFIC_PORT_SIZE;
        }

        /* return the bd to the free queue */
        PAL_cppi4QueuePush(rxfdQueueHdl,(Ptr)PAL_CPPI4_VIRT_2_PHYS((Uint32)curBD4Read), DOCSIS_MNG_QM_HDESC_SIZE_VAL, 0);
        curBD4Read = NULL;

        {
            UINT32 status;
            volatile static Int32 packet_counter = 0;

            packet_counter++;   
            PAL_cppi4Control(palHandle,PAL_CPPI41_IOCTL_GET_QUEUE_ENTRY_COUNT,&rxfdQueue, &status);
            if (debug_enable)
                printk("Read : after packet %d , free %d\n", packet_counter,status);
    } 
    } else
        offset += count; 
	
    if (debug_enable)
        printk(KERN_INFO "docsisMngRead: return %d bytes \n",count);
    return count;
}




static ssize_t docsisMngWrite(struct file *file, const char __user *buf, size_t count, loff_t *ppos)
{
    int retval;
    DocsisCppi4HostDesc *currBD = NULL;
    HalDocsisMng_t halMngData;
    struct CM_mng_t *macHdr;

    printk(KERN_INFO "Enter docsisMngWrite count %d\n", count);

    if (down_interruptible(&wr_sem))
        return -ERESTARTSYS;
    
    /* get the bd + buffer from the free queue */
    currBD = (DocsisCppi4HostDesc *) PAL_cppi4QueuePop(txfdQueueHdl);
    
    if(currBD==NULL)
    {
        printk(" %s can't allocate the free bd \n", __FUNCTION__);
        up(&wr_sem);
        return (-EAGAIN);
    }

    currBD = PAL_CPPI4_PHYS_2_VIRT(currBD);

    /* set the data pointer to copy from user */
    currBD->bufPtr = currBD->orgBufPtr + txBufferOffset;

    /*Copy the internal HAL data structure*/
    if (copy_from_user((void *)&halMngData, buf, count))
    {
        printk("%s : failed to copy HAL Mngmt data from user\n", __FUNCTION__);
        return -EFAULT;
    }

    printk(KERN_INFO "%s: encrFlag = %d msgBufLen=%d\n", __FUNCTION__, halMngData.encrFlag, halMngData.msgBufLen);

    if (halMngData.msgBufLen >= DOCSIS_MNG_TX_BUFFER_SIZE)
        return -EINVAL;     /* should be a whole message */

    /* Copy the management message buffer */
    if (copy_from_user((void *)(currBD->bufPtr), halMngData.msgBuf, halMngData.msgBufLen))
    {
        /* return the bd to the free queue */
        printk("%s : failed to copy management message buffer from user\n", __FUNCTION__);
        currBD->bufPtr = 0;
        PAL_CPPI4_CACHE_WRITEBACK(currBD, CPPI4_BD_LENGTH_FOR_CACHE);

        PAL_cppi4QueuePush (txfdQueueHdl,(Ptr) PAL_CPPI4_VIRT_2_PHYS((Uint32)currBD), 
                                                    DOCSIS_MNG_QM_HDESC_SIZE_VAL, 0);

        return -EINVAL; 
    }
    else 
    {
        /* set the bd fields and writeback the bd */
        currBD->buffLen = halMngData.msgBufLen;
        currBD->descInfo = PAL_CPPI4_HOSTDESC_DESC_TYPE_HOST << PAL_CPPI4_HOSTDESC_DESC_TYPE_SHIFT | halMngData.msgBufLen;
        
        if (!halMngData.encrFlag) 
            currBD->protocolSpec0 = US_PSI_PRIVACY_OFF_BIT_MASK;
        else
            currBD->protocolSpec0 = 0;

        macHdr = (struct CM_mng_t *)currBD->bufPtr;
        if (HAL_DocsisQosClass_ClassifyMacMngCB)
        {
            Int16 sfIndex;
            Int16 phsIndex;

            /* Classifier exceptions - don't classify */
            if(!((macHdr->type == MAC_MNGT_MSG_TYPE_RNG_REQ) ||
                 (macHdr->type == MAC_MNGT_MSG_TYPE_REG_REQ) ||
                 (macHdr->type == MAC_MNGT_MSG_TYPE_REG_RSP) ||
                 (macHdr->type == MAC_MNGT_MSG_TYPE_REG_ACK) ||
                 (macHdr->type == MAC_MNGT_MSG_TYPE_INIT_RNG_REQ)   ||
                 (macHdr->type == MAC_MNGT_MSG_TYPE_B_INIT_RNG_REQ) ||
                 (macHdr->type == MAC_MNGT_MSG_TYPE_REG_REQ_MP) ||
                 (macHdr->type == MAC_MNGT_MSG_TYPE_REG_RSP_MP) || 
                 (macHdr->type == MAC_MNGT_MSG_TYPE_DPV_RSP))) 
            {
                HAL_DocsisQosClass_ClassifyMacMngCB((Uint8 *)currBD->bufPtr, &sfIndex,  &phsIndex);
                currBD->protocolSpec0 |= ((((Uint32) sfIndex) << US_PSI_SFINDEX_SHIFT));
            }
         }

        /* DPV-RSP is sent over user-specified service flow if so requested. */
        if ((macHdr->type == MAC_MNGT_MSG_TYPE_DPV_RSP) && halMngData.forceSF)
        {
            printk(KERN_INFO "%s: Forcing SF index 0x%x for DPV-RSP\n", __FUNCTION__, halMngData.sfIndex);
            currBD->protocolSpec0 |= (halMngData.sfIndex << US_PSI_SFINDEX_SHIFT);
        }

        printk(KERN_INFO "%s: currBD->protocolSpec0 = %x currBD->buffLen=%d\n", __FUNCTION__, currBD->protocolSpec0, currBD->buffLen);

		PAL_CPPI4_CACHE_WRITEBACK(currBD, CPPI4_BD_LENGTH_FOR_CACHE);

        
        /* writeback the data */
        PAL_CPPI4_CACHE_WRITEBACK(currBD->bufPtr, halMngData.msgBufLen);

        /* push the bd to the management tx queue */
        /* return policy in the bd should be set to auto recycling */
        PAL_cppi4QueuePush (txQueueHdl,(Ptr) PAL_CPPI4_VIRT_2_PHYS((Uint32)currBD), 
                                                                DOCSIS_MNG_QM_HDESC_SIZE_VAL, halMngData.msgBufLen);
    }

    *ppos += count;
    retval = count;

    up(&wr_sem);

    return retval;
}

#ifdef POOLING
void
docsisTimerFunc(unsigned long data)
{
    UINT32 status;
    
    PAL_cppi4Control (palHandle,PAL_CPPI41_IOCTL_GET_QUEUE_PEND_STATUS,&rxQueue, &status);
    
    if (status) {
        UINT32 tmp;
        PAL_cppi4Control (palHandle,PAL_CPPI41_IOCTL_GET_QUEUE_PEND_STATUS,&rxfdQueue, &tmp);
        if (debug_enable)
            printk("Timer : waiting %d packets\n", tmp);

        if (debug_enable)
            printk(KERN_INFO "docsisTimerFunc : wake the read \n"); 
        /* wake up (unblock) for reading data from userspace   */
        dataReady = True;
        wake_up_interruptible(&docsis_mng_wait);
    }

//  mngPoolTimer.expires = jiffies + HZ;
    mngPoolTimer.expires = jiffies + HZ/100;
    add_timer(&mngPoolTimer);
}
#endif

static unsigned int
docsisMngPoll(struct file *file, poll_table * wait)
{
    if (debug_enable)
        printk("docsisMngPoll: curBD4Read %x\n", (Uint32)curBD4Read);
    poll_wait(file, &docsis_mng_wait, wait);
    if (debug_enable)
        printk("docsisMngPoll: post poll_wait curBD4Read %x\n", (Uint32)curBD4Read);
    if (curBD4Read || (curBD4Read = docsisMngGetNextBd())) {
        if (debug_enable)
            printk("docsisMngPoll: before POLLIN return curBD4Read %x\n", (Uint32)curBD4Read);
        return POLLIN | POLLRDNORM;
    }
    return 0;
}

struct file_operations docsis_mng_fops = {
    .owner   = THIS_MODULE,       
    .llseek  = NULL,              
    .read    = docsisMngRead,   
    .write   = docsisMngWrite,              
    .poll    = docsisMngPoll,  
    .ioctl   = NULL,  
    .open    = docsisMngOpen,   
    .release = docsisMngRelease,
};

static int
mng_read_proxy(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
    int len = 0;

    len += sprintf(buf + len, "debug state %d \n",debug_enable);
    debug_enable = 0;
    return(len);
}

static void
cleanup(int level)
{
    switch (level) {
    default:
    case 3:
        PAL_cppi4QueueClose(palHandle, rxQueueHdl);
    case 2:
        PAL_cppi4QueueClose(palHandle, rxfdQueueHdl);
    case 1:
        PAL_cppi4Exit(palHandle, NULL);
        break;
    }
}
 
/* initialize module (and interrupt) */
static int __init docsisMngInit (void) 
{
    int i,cnt;
    PAL_Result retCode;
    DocsisCppi4HostDesc *currBD;
    Ptr curBuffer;

    init_MUTEX(&wr_sem);
    if (debug_enable)
        printk(KERN_INFO "docsisMngInit: initializing module\n");
    
    /* get a handle to the PAL structure */
    if ((palHandle =  PAL_cppi4Init(NULL, NULL)) == NULL) {
        printk(KERN_WARNING "docsisMngInit: PAL CPPI is not initialized yet \n");
        return(-EAGAIN); /* what should i return ???*/
    }   

    printk(KERN_INFO "docsisMngInit: pal handle %x\n", (Uint32)palHandle);
    
    /*  open a queue objects one for free descriptor queue
     *  and one for rx queue 
     */ 
    if ((rxfdQueueHdl = PAL_cppi4QueueOpen(palHandle, rxfdQueue)) == NULL) {
        printk(KERN_WARNING "docsisMngInit: Failed to open fdQueue\n");
        cleanup(1);
        return -1;
    }
    if ((rxQueueHdl = PAL_cppi4QueueOpen(palHandle, rxQueue)) == NULL) {
        printk(KERN_WARNING "docsisMngInit: Failed to open rxQueueHdl\n");
        cleanup(2);
        return -1;
    }


#ifdef _USE_ACC_LIST
    /* don't forget to init the accumulator !!!! */
    chInfo.accuCfg.accChanNum = DOCSIS_MNG_ACC_RX_CHNUM;
    chInfo.accuCfg.listBase = (Ptr) PAL_CPPI4_VIRT_2_PHYS(accWrkArea); /*malloc or static allocation ???*/
    chInfo.accuCfg.maxPageEntry = DOCSIS_MNG_ACC_MAXPAGENTRY;
    chInfo.accuCfg.pacingTickCnt = DOCSIS_MNG_ACC_INTR_DELAY;
    chInfo.accuCfg.pacingMode = DOCSIS_MNG_ACC_INTR_MODE; /* delay since first new packet */
    chInfo.accuCfg.stallAvoidance = DOCSIS_MNG_ACC_STALL_AVOID;
    chInfo.accuCfg.listCountMode = DOCSIS_MNG_ACC_TERM_MODE; /* NULL Terminate Mode - The last list entry is used to store a NULL pointer record <BR> */
    chInfo.accuCfg.listEntrySize = DOCSIS_MNG_ACC_ENTRY_TYPE;
    chInfo.accuCfg.maxPageCnt = DOCSIS_MNG_ACC_PAGE_COUNT;

    PAL_osMemSet(accWrkArea,0,sizeof(accWrkArea));

    /* invalidate the page !!!*/
    accCurPage = 0;
    accDataListPtr[0] = (Uint32) accWrkArea;
    accDataListPtr[1] = (Uint32) accWrkArea + DOCSIS_MNG_ACC_MAXPAGENTRY*4 ;
    accRxQueueHdl = PAL_cppi4QueueOpen (palHandle,accRxQueue);
#endif


#ifdef POOLING
    init_timer(&mngPoolTimer);
    mngPoolTimer.function = docsisTimerFunc;
    mngPoolTimer.expires = 0;
#endif
    if (debug_enable)
        printk(KERN_INFO "Initializing the DOCSIS CPPI Resources\n");
    if (HAL_DocsisCppiInit() != 0) {
        if (debug_enable)
            printk(KERN_INFO "Failed to initialize the DOCSIS CPPI resources \n");
        return(-1);
    }


    /* ******************* init TX part of the driver ******************* */
    /*  open a queue objects one for free descriptor queue
     *  and one for tx queue 
     */ 
    if ((txfdQueueHdl = PAL_cppi4QueueOpen(palHandle, txfdQueue)) == NULL) {
        printk(KERN_WARNING "docsisMngInit: Failed to open fdQueue\n");
        cleanup(1);
        return -1;
    }
    if ((txQueueHdl = PAL_cppi4QueueOpen(palHandle, txQueue)) == NULL) {
        printk(KERN_WARNING "docsisMngInit: Failed to open rxQueueHdl\n");
        cleanup(2);
        return -1;
    }

    /*
     * get a pointer to the BD region 
     * the region is already preallocated in the PAL init so just 
     * get a pointer to the bd area 
     */
    if ((txbdAreaAddrs = PAL_cppi4AllocDesc(palHandle, 
                CNI_MGMT_CPPI4x_TX_QMGR, DOCSIS_MNG_TX_BD_NUM, DOCSIS_MNG_TX_BD_SIZE)) == NULL) 
    {
        printk(KERN_WARNING "docsisMngInit: Failed to allocate TX descriptor\n");
        cleanup(3);
        return -1;
    }

    printk(KERN_INFO "docsisMngInit: TX bd area start address %x\n",(Uint32)txbdAreaAddrs);

    /* allocate a memory for the buffers , is it better to get it one by one?*/
    if ((retCode = PAL_osMemAlloc(0, DOCSIS_MNG_TX_BUFFER_NUM * DOCSIS_MNG_TX_BUFFER_SIZE, 
                                0, (Ptr *)&txbufferAreaAddrs)) != PAL_SOK) 
    {
        printk(KERN_WARNING "docsisMngInit:Failed to allocate memory for TX buffer queue");
        cleanup(4);
        return -1;
    }
    
    if (debug_enable) printk(KERN_INFO "docsisMngInit : buffer area start address %x\n",(Uint32)txbufferAreaAddrs);
 
    PAL_osMemSet(txbufferAreaAddrs,0,DOCSIS_MNG_TX_BUFFER_NUM * DOCSIS_MNG_TX_BUFFER_SIZE);

    /* create a dedicated free bd/buffers queue for this channel */
    /* "Slice" BD's one-by-one from the chunk, allocate a buffer and token, */
    currBD = (DocsisCppi4HostDesc *)txbdAreaAddrs;
    curBuffer = txbufferAreaAddrs;

    /* need to consider the number of actually allocated buffers !!!!*/
    /* we don't really need 32 pair for management channel, do we? */
    for (cnt = 0; cnt < DOCSIS_MNG_TX_BUFFER_NUM; cnt++) 
    {
        /* Update the hardware descriptor */
        currBD->descInfo    = ( (PAL_CPPI4_HOSTDESC_DESC_TYPE_HOST     << PAL_CPPI4_HOSTDESC_DESC_TYPE_SHIFT));
        currBD->tagInfo     = 0;
        currBD->pktInfo     = 
                                (PAL_CPPI4_HOSTDESC_PKT_TYPE_GENERIC   << PAL_CPPI4_HOSTDESC_PKT_TYPE_SHIFT)
                              | (PAL_CPPI4_HOSTDESC_PKT_RETPLCY_LINKED << PAL_CPPI4_HOSTDESC_PKT_RETPLCY_SHIFT)
                              | (PAL_CPPI4_HOSTDESC_DESC_LOC_OFFCHIP   << PAL_CPPI4_HOSTDESC_DESC_LOC_SHIFT)
                              | (txfdQueue.qMgr                        << PAL_CPPI4_HOSTDESC_PKT_RETQMGR_SHIFT)
                              | (txfdQueue.qNum                        << PAL_CPPI4_HOSTDESC_PKT_RETQNUM_SHIFT);
        
        currBD->buffLen     = 0;
        currBD->bufPtr      = 0;
        currBD->nextBDPtr   = 0;
        currBD->orgBuffLen  = DOCSIS_MNG_TX_BUFFER_SIZE;
        currBD->orgBufPtr   = PAL_CPPI4_VIRT_2_PHYS(curBuffer);

        PAL_CPPI4_CACHE_WRITEBACK(currBD, CPPI4_BD_LENGTH_FOR_CACHE);

        PAL_cppi4QueuePush (txfdQueueHdl,(Ptr) PAL_CPPI4_VIRT_2_PHYS((Uint32)currBD), 
                                                    DOCSIS_MNG_QM_HDESC_SIZE_VAL, 0);
    
        currBD = (DocsisCppi4HostDesc*) ((Uint32)currBD + DOCSIS_MNG_TX_BD_SIZE);
        curBuffer = (Ptr) ((Uint32)curBuffer + DOCSIS_MNG_TX_BUFFER_SIZE);
    }

    if (debug_enable) printk(KERN_INFO "TX init done \n");
    /* ******************* Done ******************* */


    i = register_chrdev(DOCSIS_MNG_MAJOR, DOCSIS_MNG_NAME, &docsis_mng_fops);
    if (i != 0)
        return - EIO;

    create_proc_read_entry("mng", 0, NULL, mng_read_proxy, NULL);
    return 0;
}

/* close and cleanup module */
static void __exit docsisMngCleanup (void) 
{
    if (debug_enable)
        printk(KERN_INFO "cleaning up module\n");
    unregister_chrdev(DOCSIS_MNG_MAJOR, DOCSIS_MNG_NAME);

    PAL_cppi4QueueClose (palHandle, rxfdQueueHdl);
    PAL_cppi4QueueClose (palHandle, rxQueueHdl);
    /* should i support DMA teardown */
}

/***************************************************************************************************************/
/***************************************************************************************************************/
/***************************************************************************************************************/
/***************************************************************************************************************/
/*                                     Internal DDOCSIS CPPI code                                              */
/***************************************************************************************************************/
/***************************************************************************************************************/
/***************************************************************************************************************/
/***************************************************************************************************************/
static Cppi4InitCfg cppi4InitCfg_g = 
{
    .resetLine              = 0,

    /************************************************************************/
    /*                                                                      */
    /************************************************************************/

    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].queueMgrRgnBase        = (Ptr) AVALANCHE_DOCSIS_SS_QMGR_RGN_BASE, 
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descMemRgnBase         = (Ptr) AVALANCHE_DOCSIS_SS_DESCMEM_RGN_BASE,     
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].queueMgmtRgnBase       = (Ptr) AVALANCHE_DOCSIS_SS_QMGMT_RGN_BASE,
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].queueStatusRgnBase     = (Ptr) AVALANCHE_DOCSIS_SS_QSTATUS_RGN_BASE,
    
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].LinkingRAM0Base        = IO_VIRT2PHY(AVALANCHE_DOCSIS_SS_LINKING_RAM_BASE), 
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].LinkingRAM0Size        = 1024, 
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].LinkingRAM1Base        = IO_VIRT2PHY(AVALANCHE_DOCSIS_SS_LINKING_RAM_BASE+0x1000),
    
                                                                                  /* Docsis MAC DS internal memory */
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[0].base     = (Ptr) IO_VIRT2PHY( AVALANCHE_DOCSIS_SS_DS_PACKET_RAM_BASE + 0x8000 ),
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[0].szDesc   = CNI_CPPI4x_DOCSIS_DS_DESC_SIZE, 
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[0].numDesc  = CNI_CPPI4x_DOCSIS_DS_DESC_NUM,
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[0].isOnChip = 1,    

                                                                                  /* Docsis MAC US internal memory */
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[1].base     = (Ptr) IO_VIRT2PHY( AVALANCHE_DOCSIS_SS_US_PACKET_RAM_BASE ),
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[1].szDesc   = 128, 
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[1].numDesc  = 32,
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[1].isOnChip = 1,    
        
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[2].base     = 0,
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[2].szDesc   = CNI_CPPI4x_DOCSIS_US_DESC_SIZE, 
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[2].numDesc  = CNI_CPPI4x_DOCSIS_US_DESC_NUM,
    .queueMgrInfo[PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS].descRegion[2].isOnChip = 0,
    
    /************************************************************************/
    /*                                                                      */
    /************************************************************************/

    .bufMgrBase[PAL_CPPI41_BUF_MGR_PARTITION_DOCSIS]    = (CSL_BufMgr_RegsOvly) AVALANCHE_DOCSIS_SS_BMGR_BASE,

    /************************************************************************/
    /*                                                                      */
    /************************************************************************/

    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].globalCtrlBase     = (Ptr) AVALANCHE_DOCSIS_DMA0_GBLCFG_BASE,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].chCtrlStatusBase   = (Ptr) AVALANCHE_DOCSIS_DMA0_CHNCFG_BASE,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].schedCtrlBase      = (Ptr) AVALANCHE_DOCSIS_DMA0_SCHEDCFG_BASE,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].schedTableBase     = (Ptr) AVALANCHE_DOCSIS_DMA0_SCHEDTBL_BASE,
    
    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].tdFQueue.qMgr      = 0,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].tdFQueue.qNum      = 0,
  
    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].globalCtrlBase     = (Ptr) AVALANCHE_DOCSIS_DMA1_GBLCFG_BASE,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].chCtrlStatusBase   = (Ptr) AVALANCHE_DOCSIS_DMA1_CHNCFG_BASE,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].schedCtrlBase      = (Ptr) AVALANCHE_DOCSIS_DMA1_SCHEDCFG_BASE,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].schedTableBase     = (Ptr) AVALANCHE_DOCSIS_DMA1_SCHEDTBL_BASE,

    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].tdFQueue.qMgr      = PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].tdFQueue.qNum      = 59,

    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].schedTable.numEntries  = 1,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK0].schedTable.entry = 
    { 
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_TX_CH, 0 ),
    },

    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].schedTable.numEntries  = 8,
    .dmaBlock[PAL_CPPI41_DMA_BLOCK1].schedTable.entry = 
    {   
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 0 ),
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 1 ),
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 2 ),
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 3 ),
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 4 ),
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 5 ),
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 6 ), 
        PAL_CPPI41_DMA_CH_CONFIG( PAL_CPPI41_DMA_RX_CH, 7 ),
    },
};

/************************************************************************/
/*                                                                      */
/*      DOCSIS Buffer Manager specific definition                       */
/*                                                                      */
/************************************************************************/
#define DS_EMBEDDED_DESC_RAM_BASE_ADDR_HOST         IO_PHY2VIRT((0x09040000 + 0x8000)) /* Host addressing */

typedef struct
{
    Cppi4EmbdDesc   hw;
    Uint32          psi[5];
}
docsisCppiDsDescType_t;


#define DS_EMBEDDED_DESC_BPI_CMD_TEMPLATE           0x21800000
#define DS_EMBEDDED_DESC_CRC_CMD_TEMPLATE           0x41800000
#define DS_EMBEDDED_DESC_PHS_CMD_TEMPLATE           0x11800000
#define DS_EMBEDDED_DESC_RX_FIFO_0_CMD_TEMPLATE     0x72000000
#define DS_EMBEDDED_DESC_RX_FIFO_1_CMD_TEMPLATE     0x080404C1


static Int32 HAL_DocsisCppiInit(void)
{
    Cppi4BufPool        tmpBufPool; /* Used for Init calls */
    int channel;

    printk(" Entered %s \n", __FUNCTION__);

    palHandleDocsis = PAL_cppi4Init( &cppi4InitCfg_g, CPPI41_DOMAIN_PRIMARY_DOCSIS );
    
    /************************************************************************/
    /*  Buffer Pool Initialization                                          */
    /************************************************************************/
    tmpBufPool.bMgr     = PAL_CPPI41_BUF_MGR_PARTITION_DOCSIS;
    tmpBufPool.bPool    = BMGR1_POOL00;

    if ((PAL_cppi4BufPoolDirectInit(palHandleDocsis, tmpBufPool,
        BMGR1_POOL00_REF_CNT,
        BMGR1_POOL00_BUF_SIZE, 
        BMGR1_POOL00_BUF_COUNT,
        (Ptr)IO_VIRT2PHY( AVALANCHE_DOCSIS_SS_DS_PACKET_RAM_BASE ) )) == NULL)
    {
        printk ("PAL_cppi4BufPoolDirectInit for pool %d FAILED.\n", tmpBufPool.bPool);
        return -1;
    }
    /************************************************************************/


    /************************************************************************/
    /*  Format descriptors and fill in the free queue for DOCSIS Upstream   */
    /************************************************************************/
    {
        Cppi4EmbdDesc* currBD;
        int bd_cnt;
        PAL_Cppi4QueueHnd tmpQHnd; 
        Cppi4Queue tmpQ;
        
        currBD = (Cppi4EmbdDesc*)PAL_cppi4AllocDesc( palHandleDocsis, PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS, 
                                                     CNI_CPPI4x_DOCSIS_US_DESC_NUM,
                                                     CNI_CPPI4x_DOCSIS_US_DESC_SIZE );
        
        tmpQ.qMgr = PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS;
        tmpQ.qNum = CNI_CPPI4x_DOCSIS_US_FDQ;
        tmpQHnd = PAL_cppi4QueueOpen ( palHandleDocsis, tmpQ );
        
        for (bd_cnt = 0; bd_cnt < CNI_CPPI4x_DOCSIS_US_DESC_NUM; bd_cnt++)
        {
            currBD->descInfo    = CPPI41_EM_DESCINFO_DTYPE_EMBEDDED | CPPI41_EM_DESCINFO_SLOTCNT_MYCNT;
            currBD->tagInfo     = 0;
            currBD->pktInfo     = 
                 (1                << CPPI41_EM_PKTINFO_RETPOLICY_SHIFT)
                |(tmpQ.qMgr        << PAL_CPPI4_HOSTDESC_PKT_RETQMGR_SHIFT)
                |(tmpQ.qNum        << PAL_CPPI4_HOSTDESC_PKT_RETQNUM_SHIFT);
            
            PAL_CPPI4_CACHE_WRITEBACK(currBD, CNI_CPPI4x_DOCSIS_US_DESC_SIZE);
            
            PAL_cppi4QueuePush (tmpQHnd, (Ptr)PAL_CPPI4_VIRT_2_PHYS((Uint32)currBD), 
                (CNI_CPPI4x_DOCSIS_US_DESC_SIZE-24)/4, 0);
            
            currBD = (Cppi4EmbdDesc*)((Uint32)currBD + CNI_CPPI4x_DOCSIS_US_DESC_SIZE);
        }

    }
    /************************************************************************/

    
    /************************************************************************/
    /*  Format descriptors and fill in the free queue for DOCSIS Downstream */
    /************************************************************************/
    {
        docsisCppiDsDescType_t*     currBD;
        docsisCppiDsDescType_t*     currBDphysical;

        int bd_cnt;
        PAL_Cppi4QueueHnd tmpQHnd; 
        Cppi4Queue tmpQ;
        
        currBD = DS_EMBEDDED_DESC_RAM_BASE_ADDR_HOST;
        currBDphysical = (docsisCppiDsDescType_t*) IO_VIRT2PHY( AVALANCHE_DOCSIS_SS_DS_PACKET_RAM_BASE + 0x8000 );

        memset((void *)currBD, 0, CNI_CPPI4x_DOCSIS_DS_DESC_SIZE*CNI_CPPI4x_DOCSIS_DS_DESC_NUM);

        tmpQ.qMgr = PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS;
        tmpQ.qNum = CNI_CPPI4x_DOCSIS_DS_FDQ;
        tmpQHnd = PAL_cppi4QueueOpen ( palHandleDocsis, tmpQ );
        
        for (bd_cnt = 0; bd_cnt < CNI_CPPI4x_DOCSIS_DS_DESC_NUM; bd_cnt++)
        {
            currBD->hw.descInfo    = 
                  CPPI41_EM_DESCINFO_DTYPE_EMBEDDED 
                /*| CPPI41_EM_DESCINFO_SLOTCNT_MYCNT */
                | (5 << CPPI41_EM_DESCINFO_PSWSIZE_SHIFT);
            currBD->hw.tagInfo     = 0;
            currBD->hw.pktInfo     = 
                 (1                << CPPI41_EM_PKTINFO_RETPOLICY_SHIFT)
                |(1                << CPPI41_EM_PKTINFO_ONCHIP_SHIFT)
                |(tmpQ.qMgr        << PAL_CPPI4_HOSTDESC_PKT_RETQMGR_SHIFT)
                |(tmpQ.qNum        << PAL_CPPI4_HOSTDESC_PKT_RETQNUM_SHIFT);

            currBD->psi[0]  = DS_EMBEDDED_DESC_BPI_CMD_TEMPLATE;
            currBD->psi[1]  = DS_EMBEDDED_DESC_CRC_CMD_TEMPLATE;
            currBD->psi[2]  = DS_EMBEDDED_DESC_PHS_CMD_TEMPLATE;
            currBD->psi[3]  = DS_EMBEDDED_DESC_RX_FIFO_0_CMD_TEMPLATE;
            currBD->psi[4]  = DS_EMBEDDED_DESC_RX_FIFO_1_CMD_TEMPLATE;
            
            PAL_CPPI4_CACHE_WRITEBACK(currBD, CNI_CPPI4x_DOCSIS_DS_DESC_SIZE);
            
            PAL_cppi4QueuePush (tmpQHnd, (Ptr)currBDphysical, 
                (CNI_CPPI4x_DOCSIS_DS_DESC_SIZE-24)/4, 0);
            
            currBD          = (docsisCppiDsDescType_t*)((Uint32)currBD          + CNI_CPPI4x_DOCSIS_DS_DESC_SIZE);
            currBDphysical  = (docsisCppiDsDescType_t*)((Uint32)currBDphysical  + CNI_CPPI4x_DOCSIS_DS_DESC_SIZE);
        }
    }
    /************************************************************************/

    /************************************************************************/
    /*  DMA Rx Channels initialization.                                     */
    /************************************************************************/
    for (channel=0; channel<8; channel++)
    {
        Cppi4RxChInitCfg rxCh;
        PAL_Cppi4RxChHnd cppi4RxChHnd;

        /* Set up Rx channel */
        rxCh.chNum              = channel; 
        rxCh.dmaNum             = PAL_CPPI41_DMA_BLOCK1;
        rxCh.defDescType        = CPPI41_DESC_TYPE_EMBEDDED;
        rxCh.sopOffset          = 0;
        rxCh.rxCompQueue.qMgr   = PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS;
        rxCh.rxCompQueue.qNum   = 0;
        rxCh.u.embeddedPktCfg.fdQueue.qMgr = PAL_CPPI41_QUEUE_MGR_PARTITION_DOCSIS;
        rxCh.u.embeddedPktCfg.fdQueue.qNum = CNI_CPPI4x_DOCSIS_DS_FDQ;
        rxCh.u.embeddedPktCfg.numBufSlot = (EMSLOTCNT-1);
        rxCh.u.embeddedPktCfg.sopSlotNum = 0;
        rxCh.u.embeddedPktCfg.fBufPool[0].bMgr  = PAL_CPPI41_BUF_MGR_PARTITION_DOCSIS;
        rxCh.u.embeddedPktCfg.fBufPool[0].bPool = BMGR1_POOL00;
        rxCh.u.embeddedPktCfg.fBufPool[1].bMgr  = PAL_CPPI41_BUF_MGR_PARTITION_DOCSIS;
        rxCh.u.embeddedPktCfg.fBufPool[1].bPool = BMGR1_POOL00;
        rxCh.u.embeddedPktCfg.fBufPool[2].bMgr  = PAL_CPPI41_BUF_MGR_PARTITION_DOCSIS;
        rxCh.u.embeddedPktCfg.fBufPool[2].bPool = BMGR1_POOL00;
        rxCh.u.embeddedPktCfg.fBufPool[3].bMgr  = PAL_CPPI41_BUF_MGR_PARTITION_DOCSIS;
        rxCh.u.embeddedPktCfg.fBufPool[3].bPool = BMGR1_POOL00;
        cppi4RxChHnd        = PAL_cppi4RxChOpen( palHandleDocsis, &rxCh, NULL );

        /* Enable Rx channel */
        PAL_cppi4EnableRxChannel (cppi4RxChHnd, NULL);
    }
    /************************************************************************/


    /************************************************************************/
    /*  DMA Tx Channel initialization.                                      */
    /************************************************************************/
    {
        Cppi4TxChInitCfg txCh;
        PAL_Cppi4TxChHnd cppi4TxChHnd;

        /* Set up Tx channel */
        txCh.chNum          = 0;
        txCh.dmaNum         = PAL_CPPI41_DMA_BLOCK0;
        txCh.tdQueue.qMgr   = PAL_CPPI41_BUF_MGR_PARTITION_DOCSIS;
        txCh.tdQueue.qNum   = CNI_CPPI4x_DOCSIS_DS_CoP_Q;

        cppi4TxChHnd        = PAL_cppi4TxChOpen( palHandleDocsis, &txCh, NULL );

        /* Enable Rx channel */
        PAL_cppi4EnableTxChannel (cppi4TxChHnd, NULL);
    }
    /************************************************************************/

    return (0);
}



/**************************************************************************/
/*! \fn int HAL_DocsisRegisterQosClassifierMacMgmtCB(int (*QosClass_ClassifyMacMng)
 *  (Uint8 *buf, int16 *sfIndex, int16 *phsIndex)) 
 *                                      
 **************************************************************************
 *  \brief register the QOS Classifier Mac Mgmt CB function
 *  \param[in] the QOS Mac Mgmt CB function
 *  \return OK
 **************************************************************************/
int HAL_DocsisRegisterQosClassifierMacMgmtCB(int (*QosClass_ClassifyMacMng)(Uint8 *buf,Int16 *sfIndex,Int16 *phsIndex))
{   
    unsigned long flags;

    local_irq_save(flags);
    HAL_DocsisQosClass_ClassifyMacMngCB = QosClass_ClassifyMacMng;
    local_irq_restore(flags);

    return 0;
}

/***************************************************************************************************************/
/*                                  Internal DDOCSIS CPPI code - end                                           */
/***************************************************************************************************************/


EXPORT_SYMBOL(HAL_DocsisRegisterQosClassifierMacMgmtCB); 


module_init(docsisMngInit);
module_exit(docsisMngCleanup);
MODULE_AUTHOR("Texas Instruments, Inc");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Cable Modem Management Interface driver");

